@@ -17,7 +17,7 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- *  $Id: ephy-state.c,v 1.36.4.1 2005/10/02 22:35:02 chpe Exp $
+ *  $Id: ephy-state.c,v 1.39 2006/02/04 13:57:42 paharvey Exp $
  */
 
 #include "config.h"
@@ -32,6 +32,7 @@
 #include <gtk/gtkwindow.h>
 #include <gtk/gtkpaned.h>
 #include <gtk/gtkexpander.h>
+#include <gtk/gtktogglebutton.h>
 
 #define EPHY_STATES_XML_FILE	"states.xml"
 #define EPHY_STATES_XML_ROOT    (const xmlChar *)"ephy_states"
@@ -47,7 +48,7 @@ enum
 	EPHY_NODE_STATE_PROP_POSITION_Y = 7,
 	EPHY_NODE_STATE_PROP_SIZE = 8,
 	EPHY_NODE_STATE_PROP_POSITION = 9,
-	EPHY_NODE_STATE_PROP_EXPANDED = 10
+	EPHY_NODE_STATE_PROP_ACTIVE = 10
 };
 
 static EphyNode *states = NULL;
@@ -463,17 +464,30 @@ sync_expander_cb (GtkExpander *expander,
 
 	g_value_init (&value, G_TYPE_BOOLEAN);
 	g_value_set_boolean (&value, gtk_expander_get_expanded (expander));
-	ephy_node_set_property (node, EPHY_NODE_STATE_PROP_EXPANDED, &value);
+	ephy_node_set_property (node, EPHY_NODE_STATE_PROP_ACTIVE, &value);
+	g_value_unset (&value);
+}
+
+static void
+sync_toggle_cb (GtkToggleButton *toggle,
+		GParamSpec *pspec,
+		EphyNode *node)
+{
+	GValue value = { 0, };
+
+	g_value_init (&value, G_TYPE_BOOLEAN);
+	g_value_set_boolean (&value, gtk_toggle_button_get_active (toggle));
+	ephy_node_set_property (node, EPHY_NODE_STATE_PROP_ACTIVE, &value);
 	g_value_unset (&value);
 }
 
 void 
-ephy_state_add_expander	(GtkWidget *expander,
+ephy_state_add_expander (GtkWidget *widget,
 			 const char *name,
 			 gboolean default_state)
 {
 	EphyNode *node;
-	gboolean expanded;
+	gboolean active;
 
 	ensure_states ();
 
@@ -494,17 +508,25 @@ ephy_state_add_expander	(GtkWidget *expander,
 		g_value_init (&value, G_TYPE_BOOLEAN);
 		g_value_set_boolean (&value, default_state);
 		ephy_node_set_property
-			(node, EPHY_NODE_STATE_PROP_EXPANDED, &value);
+			(node, EPHY_NODE_STATE_PROP_ACTIVE, &value);
 		g_value_unset (&value);
 	}
 
-	expanded = ephy_node_get_property_boolean
-		(node, EPHY_NODE_STATE_PROP_EXPANDED);
+	active = ephy_node_get_property_boolean
+		(node, EPHY_NODE_STATE_PROP_ACTIVE);
 
-	gtk_expander_set_expanded (GTK_EXPANDER (expander), expanded);
-
-	g_signal_connect (expander, "notify::expanded",
-			  G_CALLBACK (sync_expander_cb), node);
+	if (GTK_IS_TOGGLE_BUTTON (widget))
+	{
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), active);
+		g_signal_connect (widget, "notify::active",
+				  G_CALLBACK (sync_toggle_cb), node);
+	}
+	else if (GTK_IS_EXPANDER (widget))
+	{
+		gtk_expander_set_expanded (GTK_EXPANDER (widget), active);
+		g_signal_connect (widget, "notify::expanded",
+				  G_CALLBACK (sync_expander_cb), node);
+	}
 }
 
 void
