@@ -16,7 +16,7 @@
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
- *  $Id: ephy-bookmarks.c,v 1.121 2006/01/16 23:51:41 chpe Exp $
+ *  $Id: ephy-bookmarks.c,v 1.128.2.1 2006/05/11 12:20:06 chpe Exp $
  */
 
 #include "config.h"
@@ -51,7 +51,7 @@
 #define BOOKMARKS_SAVE_DELAY (3 * 1000) /* ms */
 #define MAX_FAVORITES_NUM 10
 #define UPDATE_URI_DATA_KEY "updated-uri"
-#define SD_RESOLVE_TIMEOUT (3 * 1000) /* ms */
+#define SD_RESOLVE_TIMEOUT 0 /* ms; 0 means no timeout */
 
 #define EPHY_BOOKMARKS_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), EPHY_TYPE_BOOKMARKS, EphyBookmarksPrivate))
 
@@ -76,7 +76,7 @@ struct _EphyBookmarksPrivate
 	/* Local sites */
 	EphyNode *local;
 	GnomeVFSDNSSDBrowseHandle *browse_handle;
-	GList *resolve_list;
+	GHashTable *resolve_handles;
 #endif
 };
 
@@ -232,7 +232,7 @@ save_filter_local (EphyNode *node,
 {
 	EphyBookmarksPrivate *priv = bookmarks->priv;
 
-	return !ephy_node_has_child (priv->local, node);
+	return !priv->local || !ephy_node_has_child (priv->local, node);
 }
 #endif
 
@@ -341,41 +341,47 @@ ephy_bookmarks_update_favorites (EphyBookmarks *eb)
 }
 
 static gboolean
-add_to_favorites (EphyBookmarks *eb, EphyNode *node, EphyHistory *eh)
+add_to_favorites (EphyBookmarks *bookmarks,
+		  EphyNode *node,
+		  EphyHistory *history)
 {
+	EphyBookmarksPrivate *priv = bookmarks->priv;
 	const char *url;
 	gboolean full_menu;
 	double score;
 
-	if (ephy_node_has_child (eb->priv->favorites, node)) return FALSE;
+	if (ephy_node_db_is_immutable (priv->db)) return FALSE;
+	if (ephy_node_has_child (priv->favorites, node)) return FALSE;
 
 	url = ephy_node_get_property_string (node, EPHY_NODE_BMK_PROP_LOCATION);
-	score = get_history_item_score (eh, url);
-	full_menu = ephy_node_get_n_children (eb->priv->favorites)
+	score = get_history_item_score (history, url);
+	full_menu = ephy_node_get_n_children (priv->favorites)
 		    >= MAX_FAVORITES_NUM;
-	if (full_menu && score < eb->priv->lower_score) return FALSE;
+	if (full_menu && score < priv->lower_score) return FALSE;
 
-	if (eb->priv->lower_fav && full_menu)
+	if (priv->lower_fav && full_menu)
 	{
-		ephy_node_remove_child (eb->priv->favorites,
-					eb->priv->lower_fav);
+		ephy_node_remove_child (priv->favorites,
+					priv->lower_fav);
 	}
 
-	ephy_node_add_child (eb->priv->favorites, node);
-	ephy_bookmarks_update_favorites (eb);
+	ephy_node_add_child (priv->favorites, node);
+	ephy_bookmarks_update_favorites (bookmarks);
 
 	return TRUE;
 }
 
 static void
-history_site_visited_cb (EphyHistory *gh, const char *url, EphyBookmarks *eb)
+history_site_visited_cb (EphyHistory *history,
+			 const char *url,
+			 EphyBookmarks *bookmarks)
 {
 	EphyNode *node;
 
-	node = ephy_bookmarks_find_bookmark (eb, url);
+	node = ephy_bookmarks_find_bookmark (bookmarks, url);
 	if (node == NULL) return;
 
-	add_to_favorites (eb, node, gh);
+	add_to_favorites (bookmarks, node, history);
 }
 
 static void
@@ -408,7 +414,8 @@ clear_favorites (EphyBookmarks *bookmarks)
 }
 
 static void
-history_cleared_cb (EphyHistory *history, EphyBookmarks *bookmarks)
+history_cleared_cb (EphyHistory *history,
+		    EphyBookmarks *bookmarks)
 {
 	clear_favorites (bookmarks);
 }
@@ -659,6 +666,61 @@ update_bookmark_editing (EphyBookmarks *eb)
 		      NULL);
 }
 
+static void
+fix_hierarchy_topic (EphyBookmarks *eb,
+		     EphyNode *topic)
+{
+	GPtrArray *children;
+	EphyNode *bookmark;
+	const char *name;
+	char **split;
+	int i, j;
+
+	children = ephy_node_get_children (topic);
+	name = ephy_node_get_property_string (topic, EPHY_NODE_KEYWORD_PROP_NAME);
+	split = g_strsplit (name, "->", -1);
+	
+	for (i = 0; split[i]; i++)
+	{
+		if (split[i][0] == '\0') continue;
+		
+		topic = ephy_bookmarks_find_keyword (eb, split[i], FALSE);
+		if (topic == NULL)
+		{
+			topic = ephy_bookmarks_add_keyword (eb, split[i]);
+		}
+		for (j = 0; j < children->len; j++)
+		{
+			bookmark = g_ptr_array_index (children, j);
+			ephy_bookmarks_set_keyword (eb, topic, bookmark);
+		}
+	}
+	
+	g_strfreev (split);
+}
+
+static void
+fix_hierarchy (EphyBookmarks *eb)
+{
+	GPtrArray *topics;
+	EphyNode *topic;
+	const char *name;
+	int i;
+	
+	topics = ephy_node_get_children (eb->priv->keywords);
+	for (i = (int)topics->len - 1; i >= 0; i--)
+	{
+		topic = (EphyNode *)g_ptr_array_index (topics, i);
+		name = ephy_node_get_property_string
+			(topic, EPHY_NODE_KEYWORD_PROP_NAME);
+		if (strstr (name, "->") != NULL)
+		{
+			fix_hierarchy_topic (eb, topic);
+			ephy_node_remove_child (eb->priv->keywords, topic);
+		}
+	}
+}
+
 static void
 disable_bookmark_editing_notifier (GConfClient *client,
 				   guint cnxn_id,
@@ -703,6 +765,56 @@ backup_file (const char *original_filename, const char *extension)
 
 #ifdef ENABLE_ZEROCONF
 
+static char *
+get_id_for_service (const GnomeVFSDNSSDService *service)
+{
+	/* FIXME: limit length! */
+	return g_strdup_printf ("%s\1%s\1%s",
+				service->type,
+				service->domain,
+				service->name);
+}
+
+static EphyNode *
+get_node_for_service (EphyBookmarks *bookmarks,
+		      const GnomeVFSDNSSDService *service)
+{
+	EphyBookmarksPrivate *priv = bookmarks->priv;
+	EphyNode *kid, *node = NULL;
+	GPtrArray *children;
+	char *search_id;
+	const char *id;
+	guint i;
+
+	search_id = get_id_for_service (service);
+
+	children = ephy_node_get_children (priv->local);
+	for (i = 0; i < children->len; i++)
+	{
+		kid = g_ptr_array_index (children, i);
+
+		id = ephy_node_get_property_string (kid,
+						    EPHY_NODE_BMK_PROP_SERVICE_ID);
+
+		if (g_str_equal (id, search_id))
+		{
+			node = kid;
+			break;
+		}
+	}
+
+	g_free (search_id);
+
+	return node;
+}
+
+typedef struct
+{
+	EphyBookmarks *bookmarks;
+	EphyNode *node;
+	guint new_node : 1;
+} ResolveData;
+
 static void
 resolve_cb (GnomeVFSDNSSDResolveHandle *handle,
 	    GnomeVFSResult result,
@@ -712,17 +824,30 @@ resolve_cb (GnomeVFSDNSSDResolveHandle *handle,
 	    /* const */ GHashTable *text,
 	    int text_raw_len,
 	    const char *text_raw,
-	    EphyBookmarks *bookmarks)
+	    ResolveData *data)
 {
+	EphyBookmarks *bookmarks = data->bookmarks;
 	EphyBookmarksPrivate *priv = bookmarks->priv;
-	EphyNode *node;
+	EphyNode *node = data->node;
 	GValue value = { 0, };
 	const char *path = NULL;
 	char *url;
+	gboolean was_immutable;
 
-	priv->resolve_list = g_list_remove (priv->resolve_list, handle);
+	was_immutable = ephy_node_db_is_immutable (priv->db);
+	ephy_node_db_set_immutable (priv->db, FALSE);
 
-	if (result != GNOME_VFS_OK) return;
+	g_hash_table_steal (priv->resolve_handles, node);
+
+	/* Error, don't add the service */
+	if (result != GNOME_VFS_OK)
+	{
+		ephy_node_unref (node);
+
+		ephy_node_db_set_immutable (priv->db, was_immutable);
+
+		return;
+	}
 
 	if (text != NULL)
 	{
@@ -733,31 +858,25 @@ resolve_cb (GnomeVFSDNSSDResolveHandle *handle,
 		path = "/";
 	}
 
+	LOG ("0conf RESOLVED type=%s domain=%s name=%s => host=%s port=%d path=%s\n",
+	    service->type, service->domain, service->name,
+	    host, port, path);
+
+	/* FIXME: limit length! */
 	url = g_strdup_printf ("http://%s:%d%s", host, port, path);
 
-	node = ephy_node_new (priv->db);
-	if (node == NULL) return;
-
-	/* don't allow dragging this node */
-	ephy_node_set_is_drag_source (node, FALSE);
-
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, service->name);
-	ephy_node_set_property (node, EPHY_NODE_BMK_PROP_TITLE, &value);
-	g_value_unset (&value);
-
 	g_value_init (&value, G_TYPE_STRING);
 	g_value_take_string (&value, url);
 	ephy_node_set_property (node, EPHY_NODE_BMK_PROP_LOCATION, &value);
 	g_value_unset (&value);
 
-	g_value_init (&value, G_TYPE_BOOLEAN);
-	g_value_set_boolean (&value, TRUE);
-	ephy_node_set_property (node, EPHY_NODE_BMK_PROP_IMMUTABLE, &value);
-	g_value_unset (&value);
-
-	ephy_node_add_child (priv->bookmarks, node);
-	ephy_node_add_child (priv->local, node);
+	if (data->new_node)
+	{
+		ephy_node_add_child (priv->bookmarks, node);
+		ephy_node_add_child (priv->local, node);
+	}
+	
+	ephy_node_db_set_immutable (priv->db, was_immutable);
 }
 
 static void
@@ -768,25 +887,21 @@ browse_cb (GnomeVFSDNSSDBrowseHandle *handle,
 {
 	EphyBookmarksPrivate *priv = bookmarks->priv;
 	GnomeVFSDNSSDResolveHandle *reshandle = NULL;
-	GPtrArray *children;
-	EphyNode *kid;
-	const char *title;
-	guint i;
+	ResolveData *data;
+	EphyNode *node;
+	GValue value = { 0, };
+	gboolean new_node = FALSE;
+
+	if (service == NULL) return;
+
+	node = get_node_for_service (bookmarks, service);
 
 	if (status == GNOME_VFS_DNS_SD_SERVICE_REMOVED)
 	{
-		/* Find the bookmark to remove */
-		children = ephy_node_get_children (priv->local);
-		for (i = 0; i < children->len; i++)
+		if (node != NULL)
 		{
-			kid = g_ptr_array_index (children, i);
-			title = ephy_node_get_property_string (kid, EPHY_NODE_BMK_PROP_TITLE);
-
-			if (g_str_equal (title, service->name))
-			{
-				ephy_node_remove_child (priv->local, kid);
-				break;
-			}
+			g_hash_table_remove (priv->resolve_handles, node);
+			ephy_node_unref (node);
 		}
 
 		return;
@@ -794,16 +909,66 @@ browse_cb (GnomeVFSDNSSDBrowseHandle *handle,
 
 	/* status == GNOME_VFS_DNS_SD_SERVICE_ADDED */
 
+	LOG ("0conf ADD: type=%s domain=%s name=%s\n",
+	      service->type, service->domain, service->name);
+
+	if (node != NULL &&
+	    g_hash_table_lookup (priv->resolve_handles, node) != NULL) return;
+
+	if (node == NULL)
+	{
+		gboolean was_immutable;
+
+		was_immutable = ephy_node_db_is_immutable (priv->db);
+		ephy_node_db_set_immutable (priv->db, FALSE);
+
+		node = ephy_node_new (priv->db);
+		g_assert (node != NULL);
+
+		new_node = TRUE;
+
+		/* don't allow dragging this node */
+		ephy_node_set_is_drag_source (node, FALSE);
+
+		g_value_init (&value, G_TYPE_STRING);
+		g_value_take_string (&value, get_id_for_service (service));
+		ephy_node_set_property (node, EPHY_NODE_BMK_PROP_SERVICE_ID, &value);
+		g_value_unset (&value);
+
+		/* FIXME: limit length! */
+		g_value_init (&value, G_TYPE_STRING);
+		g_value_set_string (&value, service->name);
+		ephy_node_set_property (node, EPHY_NODE_BMK_PROP_TITLE, &value);
+		g_value_unset (&value);
+
+		g_value_init (&value, G_TYPE_BOOLEAN);
+		g_value_set_boolean (&value, TRUE);
+		ephy_node_set_property (node, EPHY_NODE_BMK_PROP_IMMUTABLE, &value);
+		g_value_unset (&value);
+
+		ephy_node_db_set_immutable (priv->db, was_immutable);
+	}
+
+	data = g_new (ResolveData, 1);
+	data->bookmarks = bookmarks;
+	data->node = node;
+	data->new_node = new_node;
+
 	if (gnome_vfs_dns_sd_resolve (&reshandle,
 				      service->name, service->type, service->domain,
 				      SD_RESOLVE_TIMEOUT,
 				      (GnomeVFSDNSSDResolveCallback) resolve_cb,
-				      bookmarks,
-				      NULL) == GNOME_VFS_OK)
+				      data,
+				      (GDestroyNotify) g_free) != GNOME_VFS_OK)
 	{
-		priv->resolve_list =
-			g_list_prepend (priv->resolve_list, reshandle);
+		ephy_node_unref (node);
+		g_free (data);
+
+		return;
 	}
+
+	g_hash_table_insert (priv->resolve_handles,
+			     node, reshandle);
 }
 
 static void
@@ -811,6 +976,11 @@ ephy_local_bookmarks_init (EphyBookmarks *bookmarks)
 {
 	EphyBookmarksPrivate *priv = bookmarks->priv;
 
+	priv->resolve_handles =
+		g_hash_table_new_full (g_direct_hash, g_direct_equal,
+				       NULL,
+				       (GDestroyNotify) gnome_vfs_dns_sd_cancel_resolve);
+
 	if (gnome_vfs_dns_sd_browse (&priv->browse_handle,
 				     "local", "_http._tcp",
 				     (GnomeVFSDNSSDBrowseCallback) browse_cb,
@@ -834,10 +1004,11 @@ ephy_local_bookmarks_stop (EphyBookmarks *bookmarks)
 		ephy_node_remove_child (priv->keywords, priv->local);
 	}
 
-	g_list_foreach (priv->resolve_list,
-			(GFunc) gnome_vfs_dns_sd_cancel_resolve, NULL);
-	g_list_free (priv->resolve_list);
-	priv->resolve_list = NULL;
+	if (priv->resolve_handles != NULL)
+	{
+		g_hash_table_destroy (priv->resolve_handles);
+		priv->resolve_handles = NULL;
+	}
 }
 
 #endif /* ENABLE_ZEROCONF */
@@ -995,6 +1166,8 @@ ephy_bookmarks_init (EphyBookmarks *eb)
 	{
 		ephy_bookmarks_init_defaults (eb);
 	}
+	
+	fix_hierarchy (eb);
 
 	eb->priv->disable_bookmark_editing_notifier_id = eel_gconf_notification_add
 		(CONF_LOCKDOWN_DISABLE_BOOKMARK_EDITING,
@@ -1009,29 +1182,30 @@ static void
 ephy_bookmarks_finalize (GObject *object)
 {
 	EphyBookmarks *eb = EPHY_BOOKMARKS (object);
+	EphyBookmarksPrivate *priv = eb->priv;
 
-	eel_gconf_notification_remove (eb->priv->disable_bookmark_editing_notifier_id);
+	eel_gconf_notification_remove (priv->disable_bookmark_editing_notifier_id);
 
-	if (eb->priv->save_timeout_id != 0)
+	if (priv->save_timeout_id != 0)
 	{
-		g_source_remove (eb->priv->save_timeout_id);
+		g_source_remove (priv->save_timeout_id);
 	}
 
+	ephy_bookmarks_save (eb);
+
 #ifdef ENABLE_ZEROCONF
 	ephy_local_bookmarks_stop (eb);
 #endif
 
-	ephy_bookmarks_save (eb);
+	ephy_node_unref (priv->bookmarks);
+	ephy_node_unref (priv->keywords);
+	ephy_node_unref (priv->favorites);
+	ephy_node_unref (priv->notcategorized);
 
-	ephy_node_unref (eb->priv->bookmarks);
-	ephy_node_unref (eb->priv->keywords);
-	ephy_node_unref (eb->priv->favorites);
-	ephy_node_unref (eb->priv->notcategorized);
+	g_object_unref (priv->db);
 
-	g_object_unref (eb->priv->db);
-
-	g_free (eb->priv->xml_file);
-	g_free (eb->priv->rdf_file);
+	g_free (priv->xml_file);
+	g_free (priv->rdf_file);
 
 	LOG ("Bookmarks finalized");
 
@@ -1041,11 +1215,7 @@ ephy_bookmarks_finalize (GObject *object)
 EphyBookmarks *
 ephy_bookmarks_new (void)
 {
-	EphyBookmarks *eb;
-
-	eb = EPHY_BOOKMARKS (g_object_new (EPHY_TYPE_BOOKMARKS, NULL));
-
-	return eb;
+	return EPHY_BOOKMARKS (g_object_new (EPHY_TYPE_BOOKMARKS, NULL));
 }
 
 static void
@@ -1186,6 +1356,85 @@ ephy_bookmarks_find_bookmark (EphyBookmarks *eb,
 	return NULL;
 }
 
+static gboolean
+is_similar (const char *url1, const char *url2)
+{
+	int i, j;
+	
+	for (i = 0; url1[i]; i++)
+	  if (url1[i] == '#' || url1[i] == '?')
+	    break;
+	while(i>0 && url1[i] == '/')
+	  i--;
+	
+	for (j = 0; url2[j]; j++)
+	  if (url2[j] == '#' || url2[j] == '?')
+	    break;
+	while(j>0 && url2[j] == '/') 
+	  j--;
+	
+	if (i != j) return FALSE;
+	if (strncmp (url1, url2, i) != 0) return FALSE;
+	return TRUE;
+}
+
+gint
+ephy_bookmarks_get_similar (EphyBookmarks *eb,
+			    EphyNode *bookmark,
+			    GPtrArray *identical,
+			    GPtrArray *similar)
+{
+	GPtrArray *children;
+	const char *url;
+	int i, result;
+
+	g_return_val_if_fail (EPHY_IS_BOOKMARKS (eb), -1);
+	g_return_val_if_fail (eb->priv->bookmarks != NULL, -1);
+	g_return_val_if_fail (bookmark != NULL, -1);
+	
+	url = ephy_node_get_property_string 
+	  (bookmark, EPHY_NODE_BMK_PROP_LOCATION);
+
+	g_return_val_if_fail (url != NULL, -1);
+	
+	result = 0;
+	
+	children = ephy_node_get_children (eb->priv->bookmarks);
+	for (i = 0; i < children->len; i++)
+	{
+		EphyNode *kid;
+		const char *location;
+
+		kid = g_ptr_array_index (children, i);
+		if (kid == bookmark)
+		{
+			continue;
+		}
+		
+		location = ephy_node_get_property_string
+			(kid, EPHY_NODE_BMK_PROP_LOCATION);
+
+		if (location != NULL)
+		{
+			if(identical != NULL && strcmp (url, location) == 0)
+			{
+				g_ptr_array_add (identical, kid);
+				result++;
+			}
+			else if(is_similar (url, location))
+			{
+				if (similar != NULL)
+				{
+					g_ptr_array_add (similar, kid);
+				}
+				result++;
+			}
+		}
+	}
+	
+	return result;
+}
+
 void
 ephy_bookmarks_set_icon	(EphyBookmarks *eb,
 			 const char *url,
